﻿using System;

using Microsoft.Extensions.Logging;

using Platform.Modules.Abstract;
using System.Threading;
using System.Collections.Concurrent;

namespace Platform.Modules
{
    public abstract class ModuleTask : IModuleTask
    {
        #region Construction and Fields

        /// <summary>
        /// Mutex Object used for synchronising code acces
        /// </summary>
        protected object Mutex = new object();

        /// <summary>
        /// Create action
        /// </summary>
        private BlockingCollection<Action> _Actions;

        /// <summary>
        /// Gets the logging interface
        /// </summary>
        protected readonly ILogger Logger;
        protected readonly ILoggerFactory LoggerFactory;

        /// <summary>
        /// Construct a new sensara worker
        /// </summary>
        /// <param name="settings"></param>
        public ModuleTask( ILoggerFactory loggerFactory = null )
        {
            // create logger
            LoggerFactory = loggerFactory;
            Logger = LoggerFactory?.CreateLogger(GetType());

            // service starting
            Logger?.LogDebug(new EventId(1, "TaskConstruction"), "{0} Task Construction", ServiceName );
        }

        /// <summary>
        /// Gets the service name
        /// </summary>
        protected abstract string ServiceName { get; }

        /// <summary>
        /// Default task is enabled
        /// </summary>
        protected virtual bool IsEnabled { get { return true; } }

        /// <summary>
        /// Gets the cancellation token source
        /// </summary>
        protected CancellationTokenSource CancellationTokenSource { get; private set; }

        /// <summary>
        /// Gets the worker thread
        /// </summary>
        protected Thread Thread { get; private set; }

        #endregion

        #region Start and Stop

        /// <summary>
        /// Returns true if worker started
        /// </summary>
        public bool IsStarted
        {
            get { return CancellationTokenSource != null; }
        }

        /// <summary>
        /// Start sensara worker
        /// </summary>
        public void Start()
        {
            lock (Mutex)
            {
                try
                {
                    // check if enabled
                    if (!IsEnabled)
                        return;

                    // allready started
                    if (IsStarted)
                        return;

                    // service starting
                    Logger?.LogInformation(new EventId(2, "TaskStarting"), "{0} Task Starting ...", ServiceName);

                    // support cancellation
                    CancellationTokenSource = new CancellationTokenSource();

                    // on starting
                    OnStart();

                    // service starting
                    Logger?.LogInformation(new EventId(3, "TaskStarted"), "{0} Task Started", ServiceName );
                }
                catch (Exception ex)
                {
                    // service starting
                    Logger?.LogError(new EventId(4, "TaskStartFailed"), ex, "{0} Task Start Failed", ServiceName );

                    // set cancel flag
                    if (CancellationTokenSource != null)
                    {
                        // set cancel flag
                        try { CancellationTokenSource.Cancel(); }
                        catch (Exception) { Logger?.LogError(new EventId(5, "TaskCleanupFailedCancel"), ex, "{0} Task cleanup Failed", ServiceName); }

                        // process stop
                        try { OnStop(); }
                        catch (Exception) { Logger?.LogError(new EventId(6, "TaskCleanupFailedStop"), ex, "{0} Task cleanup Failed", ServiceName); }

                        // cleanup cancellation flag
                        try
                        {
                            CancellationTokenSource.Dispose();
                            CancellationTokenSource = null;
                        }
                        catch (Exception) { Logger?.LogError(new EventId(7, "TaskCleanupFailedDispose"), ex, "{0} Task cleanup Failed", ServiceName); }
                    }

                    // rethrow
                    throw;
                }
            }
        }

        /// <summary>
        /// Stop sensara worker
        /// </summary>
        public void Stop()
        {
            lock (Mutex)
            {
                try
                {
                    // check thread
                    if (!IsStarted)
                        return;

                    // service starting
                    Logger?.LogInformation(new EventId(8, "TaskStopping"), "{0} Task Stopping ...", ServiceName);

                    // set cancellation
                    CancellationTokenSource.Cancel();

                    // handle stop
                    OnStop();

                    // dispose cancellation
                    CancellationTokenSource.Dispose();
                    CancellationTokenSource = null;

                    // service starting
                    Logger?.LogInformation(new EventId(9, "TaskStopped"), "{0} Task Stopped", ServiceName);
                }
                catch (Exception ex)
                {
                    // service starting
                    Logger?.LogError(new EventId(10, "TaskStopFailed"), ex, "{0} Task Stop Failed", ServiceName);

                    // rethrow
                    throw;
                }
            }
        }

        /// <summary>
        /// Restart
        /// </summary>
        public void Restart( Action callback = null )
        {
            lock (this)
            {
                try
                {
                    // if bus is started
                    if (!IsStarted)
                        return;

                    // logging
                    Logger?.LogInformation(new EventId(1, "TaskRestarting"), "{0} Task Restarting ...", ServiceName);

                    // set cancellation
                    CancellationTokenSource.Cancel();

                    // stop now
                    OnRestop();

                    // dispose cancellation
                    CancellationTokenSource.Dispose();

                    // invoke optional callback
                    callback?.Invoke();

                    // check if enabled
                    if (!IsEnabled)
                        return;

                    // support cancellation
                    CancellationTokenSource = new CancellationTokenSource();

                    // stop now
                    OnRestart();

                    // logging
                    Logger?.LogInformation(new EventId(2, "TaskRestarted"), "{0} Task Restarted", ServiceName);
                }
                catch (Exception ex)
                {
                    // logging
                    Logger?.LogError(new EventId(3, "TaskRestartFailed" ), ex, "{0} Task Restart Failed", ServiceName);

                    // force stop
                    Stop();

                    // signal EMERGENCY (email etc)
                }
            }
        }

        #endregion

        #region Default Start/Stop implementation

        /// <summary>
        /// Process start
        /// </summary>
        protected virtual void OnStart()
        {
            StartActions();
            StartThread();
        }

        /// <summary>
        /// Process stop
        /// </summary>
        protected virtual void OnStop()
        {
            StopThread();
            StopActions();
        }

        /// <summary>
        /// Re-stop implementation
        /// </summary>
        protected virtual void OnRestop()
        {
            OnStop();
        }

        /// <summary>
        /// Restart
        /// </summary>
        protected virtual void OnRestart()
        {
            OnStart();
        }

        #endregion

        #region Thread Worker

        /// <summary>
        /// Start Thread
        /// </summary>
        protected void StartThread()
        {
            // create thread
            Thread = new Thread(Worker);
            Thread.Name = ServiceName;
            Thread.Start();
        }

        /// <summary>
        /// Stop Thread
        /// </summary>
        protected void StopThread()
        {
            if (Thread != null)
            {
                Thread.Join();
                Thread = null;
            }
        }

        /// <summary>
        /// Actual background worker
        /// </summary>
        protected virtual void Worker()
        {
            // service starting
            Logger?.LogInformation(new EventId(14, "TaskWorkerStarted"), "{0} Task Worker Started", ServiceName);

            // keep running
            while (true)
            {
                try
                {
                    // single worker task
                    WorkerTask();
                }
                catch( OperationCanceledException )
                {
                    // not an error
                }
                catch (Exception ex)
                {
                    // log error
                    if (!CancellationTokenSource.IsCancellationRequested)
                        Logger?.LogError(new EventId(15, "TaskWorkerError"), ex, "{0} Task Worker Error", ServiceName);
                }

                // check if stopped
                if (CancellationTokenSource.Token.WaitHandle.WaitOne(10000))
                    break;
            }

            // service starting
            Logger?.LogInformation(new EventId(16, "TaskWorkerStopped"), "{0} Task Worker Stopped", ServiceName);
        }

        /// <summary>
        /// Implement worker task
        /// </summary>
        protected virtual void WorkerTask()
        {
            RunActions();
        }

        #endregion

        #region Action Processing

        /// <summary>
        /// Start actions
        /// </summary>
        protected void StartActions()
        {
            _Actions = new BlockingCollection<Action>();
        }

        /// <summary>
        /// Stop actions
        /// </summary>
        protected void StopActions()
        {
            if (_Actions != null)
            {
                _Actions.Dispose();
                _Actions = null;
            }
        }

        /// <summary>
        /// Add action to collection
        /// </summary>
        /// <param name="action"></param>
        protected void AddAction( Action action )
        {
            // check if enabled
            if (IsEnabled)
                _Actions.Add(action, CancellationTokenSource.Token);
        }

        /// <summary>
        /// Get idle action
        /// </summary>
        /// <param name="action"></param>
        /// <returns></returns>
        protected virtual int GetIdleAction(ref Action action)
        {
            return -1;
        }

        /// <summary>
        /// Run actions
        /// </summary>
        protected void RunActions()
        {
            // get cancellation token
            var cancellation = CancellationTokenSource.Token;

            // keep running actions
            while ( !cancellation.IsCancellationRequested )
            {
                // get idle action
                Action idleaction = null;
                int timeout = GetIdleAction(ref idleaction);

                // run scheduled action
                Action taskaction = null;
                if (_Actions.TryTake(out taskaction, timeout, cancellation)) taskaction?.Invoke();
                else if (cancellation.IsCancellationRequested) break;
                else idleaction?.Invoke();
            }
        }

        #endregion
    }
}
